﻿using System;
using System.Linq;
using System.Linq.Expressions;
using System.Collections.Generic;

namespace Achiral.Extension
{
    public static partial class MakeWorld
    {
        public static Expression<Action> Bind1st<T>(this Expression<Action<T>> expr, T arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T));
            return Expression.Lambda<Action>(Expression.Invoke(expr, a1));
        }
        public static Expression<Action<T2>> Bind1st<T1, T2>(this Expression<Action<T1, T2>> expr, T1 arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T1));
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            return Expression.Lambda<Action<T2>>(Expression.Invoke(expr, a1, a2), a2);
        }
        public static Expression<Action<T1>> Bind2nd<T1, T2>(this Expression<Action<T1, T2>> expr, T2 arg2)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Constant(arg2, typeof(T2));
            return Expression.Lambda<Action<T1>>(Expression.Invoke(expr, a1, a2), a1);
        }
        public static Expression<Action<T2, T3>> Bind1st<T1, T2, T3>(this Expression<Action<T1, T2, T3>> expr, T1 arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T1));
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            return Expression.Lambda<Action<T2, T3>>(Expression.Invoke(expr, a1, a2, a3), a2, a3);
        }
        public static Expression<Action<T1, T3>> Bind2nd<T1, T2, T3>(this Expression<Action<T1, T2, T3>> expr, T2 arg2)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Constant(arg2, typeof(T2));
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            return Expression.Lambda<Action<T1, T3>>(Expression.Invoke(expr, a1, a2, a3), a1, a3);
        }
        public static Expression<Action<T1, T2>> Bind3rd<T1, T2, T3>(this Expression<Action<T1, T2, T3>> expr, T3 arg3)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Constant(arg3, typeof(T3));
            return Expression.Lambda<Action<T1, T2>>(Expression.Invoke(expr, a1, a2, a3), a1, a2);
        }
        public static Expression<Action<T2, T3, T4>> Bind1st<T1, T2, T3, T4>(this Expression<Action<T1, T2, T3, T4>> expr, T1 arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T1));
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            var a4 = Expression.Parameter(typeof(T4), "arg4");
            return Expression.Lambda<Action<T2, T3, T4>>(Expression.Invoke(expr, a1, a2, a3, a4), a2, a3, a4);
        }
        public static Expression<Action<T1, T3, T4>> Bind2nd<T1, T2, T3, T4>(this Expression<Action<T1, T2, T3, T4>> expr, T2 arg2)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Constant(arg2, typeof(T2));
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            var a4 = Expression.Parameter(typeof(T4), "arg4");
            return Expression.Lambda<Action<T1, T3, T4>>(Expression.Invoke(expr, a1, a2, a3, a4), a1, a3, a4);
        }
        public static Expression<Action<T1, T2, T4>> Bind3rd<T1, T2, T3, T4>(this Expression<Action<T1, T2, T3, T4>> expr, T3 arg3)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Constant(arg3, typeof(T3));
            var a4 = Expression.Parameter(typeof(T4), "arg4");
            return Expression.Lambda<Action<T1, T2, T4>>(Expression.Invoke(expr, a1, a2, a3, a4), a1, a2, a4);
        }
        public static Expression<Action<T1, T2, T3>> Bind4th<T1, T2, T3, T4>(this Expression<Action<T1, T2, T3, T4>> expr, T4 arg4)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            var a4 = Expression.Constant(arg4, typeof(T4));
            return Expression.Lambda<Action<T1, T2, T3>>(Expression.Invoke(expr, a1, a2, a3, a4), a1, a2, a3);
        }

        public static Expression<Func<TResult>> Bind1st<T, TResult>(this Expression<Func<T, TResult>> expr, T arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T));
            return Expression.Lambda<Func<TResult>>(Expression.Invoke(expr, a1));
        }
        public static Expression<Func<T2, TResult>> Bind1st<T1, T2, TResult>(this Expression<Func<T1, T2, TResult>> expr, T1 arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T1));
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            return Expression.Lambda<Func<T2, TResult>>(Expression.Invoke(expr, a1, a2), a2);
        }
        public static Expression<Func<T1, TResult>> Bind2nd<T1, T2, TResult>(this Expression<Func<T1, T2, TResult>> expr, T2 arg2)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Constant(arg2, typeof(T2));
            return Expression.Lambda<Func<T1, TResult>>(Expression.Invoke(expr, a1, a2), a1);
        }
        public static Expression<Func<T2, T3, TResult>> Bind1st<T1, T2, T3, TResult>(this Expression<Func<T1, T2, T3, TResult>> expr, T1 arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T1));
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            return Expression.Lambda<Func<T2, T3, TResult>>(Expression.Invoke(expr, a1, a2, a3), a2, a3);
        }
        public static Expression<Func<T1, T3, TResult>> Bind2nd<T1, T2, T3, TResult>(this Expression<Func<T1, T2, T3, TResult>> expr, T2 arg2)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Constant(arg2, typeof(T2));
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            return Expression.Lambda<Func<T1, T3, TResult>>(Expression.Invoke(expr, a1, a2, a3), a1, a3);
        }
        public static Expression<Func<T1, T2, TResult>> Bind3rd<T1, T2, T3, TResult>(this Expression<Func<T1, T2, T3, TResult>> expr, T3 arg3)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Constant(arg3, typeof(T3));
            return Expression.Lambda<Func<T1, T2, TResult>>(Expression.Invoke(expr, a1, a2, a3), a1, a2);
        }
        public static Expression<Func<T2, T3, T4, TResult>> Bind1st<T1, T2, T3, T4, TResult>(this Expression<Func<T1, T2, T3, T4, TResult>> expr, T1 arg1)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Constant(arg1, typeof(T1));
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            var a4 = Expression.Parameter(typeof(T4), "arg4");
            return Expression.Lambda<Func<T2, T3, T4, TResult>>(Expression.Invoke(expr, a1, a2, a3, a4), a2, a3, a4);
        }
        public static Expression<Func<T1, T3, T4, TResult>> Bind2nd<T1, T2, T3, T4, TResult>(this Expression<Func<T1, T2, T3, T4, TResult>> expr, T2 arg2)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Constant(arg2, typeof(T2));
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            var a4 = Expression.Parameter(typeof(T4), "arg4");
            return Expression.Lambda<Func<T1, T3, T4, TResult>>(Expression.Invoke(expr, a1, a2, a3, a4), a1, a3, a4);
        }
        public static Expression<Func<T1, T2, T4, TResult>> Bind3rd<T1, T2, T3, T4, TResult>(this Expression<Func<T1, T2, T3, T4, TResult>> expr, T3 arg3)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Constant(arg3, typeof(T3));
            var a4 = Expression.Parameter(typeof(T4), "arg4");
            return Expression.Lambda<Func<T1, T2, T4, TResult>>(Expression.Invoke(expr, a1, a2, a3, a4), a1, a2, a4);
        }
        public static Expression<Func<T1, T2, T3, TResult>> Bind4th<T1, T2, T3, T4, TResult>(this Expression<Func<T1, T2, T3, T4, TResult>> expr, T4 arg4)
        {
            AssertNonNullArgument(expr, "expr");

            var a1 = Expression.Parameter(typeof(T1), "arg1");
            var a2 = Expression.Parameter(typeof(T2), "arg2");
            var a3 = Expression.Parameter(typeof(T3), "arg3");
            var a4 = Expression.Constant(arg4, typeof(T4));
            return Expression.Lambda<Func<T1, T2, T3, TResult>>(Expression.Invoke(expr, a1, a2, a3, a4), a1, a2, a3);
        }
    }
}