﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Achiral;
using Achiral.Extension;

namespace Achiral.Test
{
    /// <summary>
    /// Summary description for TupleTest
    /// </summary>
    [TestClass]
    public class EnumerableTest
    {
        public EnumerableTest()
        {
            //
            // TODO: Add constructor logic here
            //
        }

        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext { get; set; }

 
        #region Additional test attributes
        //
        // You can use the following additional attributes as you write your tests:
        //
        // Use ClassInitialize to run code before running the first test in the class
        // [ClassInitialize()]
        // public static void MyClassInitialize(TestContext testContext) { }
        //
        // Use ClassCleanup to run code after all tests in a class have run
        // [ClassCleanup()]
        // public static void MyClassCleanup() { }
        //
        // Use TestInitialize to run code before running each test 
        // [TestInitialize()]
        // public void MyTestInitialize() { }
        //
        // Use TestCleanup to run code after each test has run
        // [TestCleanup()]
        // public void MyTestCleanup() { }
        //
        #endregion

        [TestMethod]
        public void Select_ArrayTest1()
        {
            new[,] { { "(0,0)", "(0,1)" }, { "(1,0)", "(1,1)" }, { "(2,0)", "(2,1)" } }
                .Select((c, i0, i1) => c == string.Format("({0},{1})", i0, i1))
                .ForEach(item => Assert.IsTrue(item));

            new[, ,]
            {
                {
                    { "(0,0,0)", "(0,0,1)", "(0,0,2)" },
                    { "(0,1,0)", "(0,1,1)", "(0,1,2)" },
                },
                {
                    { "(1,0,0)", "(1,0,1)", "(1,0,2)" },
                    { "(1,1,0)", "(1,1,1)", "(1,1,2)" },
                },
            }.Select((c, i0, i1, i2) => c == string.Format("({0},{1},{2})", i0, i1, i2))
                .ForEach(item => Assert.IsTrue(item));
        }

        [TestMethod]
        public void LongSelect_ArrayTest1()
        {
            new[,] { { "(0,0)", "(0,1)" }, { "(1,0)", "(1,1)" }, { "(2,0)", "(2,1)" } }
                .LongSelect((c, i0, i1) => c == string.Format("({0},{1})", i0, i1))
                .ForEach(item => Assert.IsTrue(item));

            new[, ,]
            {
                {
                    { "(0,0,0)", "(0,0,1)", "(0,0,2)" },
                    { "(0,1,0)", "(0,1,1)", "(0,1,2)" },
                },
                {
                    { "(1,0,0)", "(1,0,1)", "(1,0,2)" },
                    { "(1,1,0)", "(1,1,1)", "(1,1,2)" },
                },
            }.LongSelect((c, i0, i1, i2) => c == string.Format("({0},{1},{2})", i0, i1, i2))
                .ForEach(item => Assert.IsTrue(item));
        }

        [TestMethod]
        public void CascadeDepthFirst_OrderTest1()
        {
            // Build up simple tree
            var lookup = new Dictionary<string, IEnumerable<string>>
            {
                { "DriveRoot", new []{"Program Files", "Users", "Windows"} },
                { "Program Files", new []{"Microsoft Visual Studio 9.0"} },
                { "Users", new []{"Alice", "Bob", "Carol"} },
                { "Bob", new []{"Project"} },
                { "Windows", new []{"System32"} },
            }.ToLookup();

            var q1 = Make.Sequence("DriveRoot").CascadeDepthFirst(lookup.AsFunc());

            var q2 = new[]
            {
                "DriveRoot",
                    "Program Files",
                        "Microsoft Visual Studio 9.0",
                    "Users",
                        "Alice",
                        "Bob",
                            "Project",
                        "Carol",
                    "Windows",
                        "System32",
            };

            Assert.IsTrue(Enumerable.SequenceEqual(q1, q2));
        }

        [TestMethod]
        public void CascadeDepthFirst_OrderTest2()
        {
            // Build up simple tree
            var lookup = new Dictionary<string, IEnumerable<string>>
            {
                { "DriveRoot", new []{"Program Files", "Users", "Windows"} },
                { "Program Files", new []{"Microsoft Visual Studio 9.0"} },
                { "Users", new []{"Alice", "Bob", "Carol"} },
                { "Bob", new []{"Project"} },
                { "Windows", new []{"System32"} },
            }.ToLookup();

            var q1 = Make.Sequence("DriveRoot").CascadeDepthFirst(lookup.AsFunc(), (dir, level) => Make.Tuple(level, dir));

            var q2 = new TupleList<int, string>()
            {
                {0, "DriveRoot"},
                    {1, "Program Files"},
                        {2, "Microsoft Visual Studio 9.0"},
                    {1, "Users"},
                        {2, "Alice"},
                        {2, "Bob"},
                            {3, "Project"},
                        {2, "Carol"},
                    {1, "Windows"},
                        {2, "System32"},
            };

            Assert.IsTrue(Enumerable.SequenceEqual(q1, q2));
        }

        [TestMethod]
        public void CascadeDepthFirst_RecursiveTest1()
        {
            const int limit = 1000000;
            var q1 = Make.Sequence(0)
                         .CascadeDepthFirst(i => i+1 < limit ? Make.Sequence(i+1) : Enumerable.Empty<int>());

            var q2 = Enumerable.Range(0, limit);

            Assert.IsTrue(Enumerable.SequenceEqual(q1, q2));
        }

        [TestMethod]
        public void CascadeBreadthFirst_ResultOrderTest1()
        {
            // Build up simple tree
            var lookup = new Dictionary<string, IEnumerable<string>>
            {
                { "DriveRoot", new []{"Program Files", "Users", "Windows"} },
                { "Program Files", new []{"Microsoft Visual Studio 9.0"} },
                { "Users", new []{"Alice", "Bob", "Carol"} },
                { "Bob", new []{"Project"} },
                { "Windows", new []{"System32"} },
            }.ToLookup();

            var q1 = Make.Sequence("DriveRoot").CascadeBreadthFirst(lookup.AsFunc());

            var q2 = new[]
            {
                "DriveRoot",
                    "Program Files",
                    "Users",
                    "Windows",
                        "Microsoft Visual Studio 9.0",
                        "Alice",
                        "Bob",
                        "Carol",
                        "System32",
                            "Project",
            };
            q1.ConsoleWriteLine();

            Assert.IsTrue(Enumerable.SequenceEqual(q1, q2));
        }

        [TestMethod]
        public void CascadeBreadthFirst_ResultOrderTest2()
        {
            // Build up simple tree
            var lookup = new Dictionary<string, IEnumerable<string>>
            {
                { "DriveRoot", new []{"Program Files", "Users", "Windows"} },
                { "Program Files", new []{"Microsoft Visual Studio 9.0"} },
                { "Users", new []{"Alice", "Bob", "Carol"} },
                { "Bob", new []{"Project"} },
                { "Windows", new []{"System32"} },
            }.ToLookup();

            var q1 = Make.Sequence("DriveRoot").CascadeBreadthFirst(lookup.AsFunc(), (dir, level) => Make.Tuple(level, dir));

            var q2 = new TupleList<int, string>()
            {
                {0, "DriveRoot"},
                    {1, "Program Files"},
                    {1, "Users"},
                    {1, "Windows"},
                        {2, "Microsoft Visual Studio 9.0"},
                        {2, "Alice"},
                        {2, "Bob"},
                        {2, "Carol"},
                        {2, "System32"},
                            {3, "Project"},
            };
            q1.ConsoleWriteLine();

            Assert.IsTrue(Enumerable.SequenceEqual(q1, q2));
        }

        [TestMethod]
        public void CascadeBreadthFirst_RecursiveTest()
        {
            var duration = TimeSpan.FromSeconds(10.0);

            // this type of query should have no limitation of recursion count
            var q1 = Make.Sequence(Make.Tuple(0, DateTime.Now))
                         .CascadeBreadthFirst(arg => Make.Sequence(Make.Tuple(arg.Arg1 + 1, arg.Arg2)))
                         .TakeWhile(arg => DateTime.Now < arg.Arg2.Add(duration));

            var count = q1.LongCount();
            Assert.Inconclusive("This is not an error, Result count = {0}", count);
        }

        [TestMethod]
        public void ForEach_SimpleTest()
        {
            const int count = 1000;
            Enumerable.Range(0, count)
                      .ForEach((item, index) => Assert.AreEqual(item, index));
        }

        [TestMethod]
        public void ZipWith_SimpleTest1()
        {
            const int count = 1000;
            Enumerable.Range(0, count)
                      .ZipWith( Enumerable.Range(1, count), (arg1, arg2) => Make.Tuple(arg1, arg2) )
                      .ForEach(arg => Assert.AreEqual(arg.Arg1 + 1, arg.Arg2));
        }

        [TestMethod]
        public void ZipWith_SimpleTest2()
        {
            var count = 1000;
            Enumerable.Range(0, count)
                      .ZipWith(Enumerable.Range(1, count), (arg1, arg2, index) => Make.Tuple(arg1, arg2, index))
                      .Select((arg, index) => Make.Tuple(arg.Arg3, index))
                      .ForEach(arg => Assert.AreEqual(arg.Arg1, arg.Arg2));
        }

        [TestMethod]
        public void ZipWith_TruncateTest()
        {
            const int count1 = 1000;
            const int count2 = 100;
            var zippedCount = 
                Enumerable.Range(0, count1)
                          .ZipWith(Enumerable.Range(1, count2), (arg1, arg2) => Make.Tuple(arg1, arg2))
                          .Count();

            Assert.AreEqual(Math.Min(count1, count2), zippedCount);
        }

        [TestMethod]
        public void ZipWith_DisposedTest()
        {
            const int count1 = 1000;
            const int count2 = 100;
            bool disposed1 = false;
            bool disposed2 = false;

            var range1 = Enumerable.Range(0, count1).Hook(
                new EnumeratorHook<int>
                {
                    OnAfterDispose = e => disposed1 = true,
                });

            var range2 = Enumerable.Range(0, count2).Hook(
                new EnumeratorHook<int>
                {
                    OnAfterDispose = e => disposed2 = true,
                });

            var zippedCount = range1.ZipWith(range2, (arg1, arg2) => Make.Tuple(arg1, arg2)).Count();

            Assert.AreEqual(true, disposed1);
            Assert.AreEqual(true, disposed2);
        }

        [TestMethod]
        public void ZipWith_InnerProduct()
        {
            var dp1 = new[] { 1, 2, 3, 4 }.ZipWith(new[] { 10, 11, 4 }, (a, b) => a * b)
                                          .Sum();
            var dp2 = new[] { 1, 2, 3, 4 }.ZipWith(new[] { 10, 11, 4 }, (a, b) => a * b)
                                          .Aggregate((x, y) => x + y);

            Assert.AreEqual(dp1, 44);
            Assert.AreEqual(dp2, 44);
        }

        [TestMethod]
        public void SelectWithFirst_SimpleTest()
        {
            const int count = 1000;
            var q1 = Enumerable.Range(0, count).SelectWithFirst((first, item) => Make.Tuple(first, item));
            var q2 = Enumerable.Range(0, count).Skip(1).Select(item => Make.Tuple(0, item));

            Assert.IsTrue(Enumerable.SequenceEqual(q1, q2));
        }

        [TestMethod]
        public void SelectWithFirst_TruncateTest()
        {
            var count = 1;
            var q = Enumerable.Range(0, count).SelectWithFirst((first, item) => Make.Tuple(first, item));

            Assert.AreEqual(0, q.Count());
        }
    }
}
